import regex
import base64
import js2py
from typing import Any

# Thanks to http://regex.inginf.units.it/
arrowexp = regex.compile(r'\w[^\}]*+')

def to_ecma5_function(f:str) -> str:
	match = arrowexp.search(f)
	conv = '(function(){' + match.group(0) + '})()'
	return regex.sub(
		r'(?:s|\(s\)) => s.split\([\'"]{2}\).reverse\(\).join\([\'"]{2}\)',
		'function(s){return s.split(\'\').reverse().join(\'\')}',
		conv
	)

def atob(s:str) -> str:
	return base64.standard_b64decode(str(s)).decode('utf-8')

def exec(f:str) -> Any:
	ctx = js2py.EvalJs({'atob': atob})
	ctx.execute('window.document = { };')
	ctx.execute('window.Map = function(_i){ };')
	ctx.execute('window.setTimeout = function(_f,_t){ };')
	ctx.execute('window.setInterval = function(_f,_t){ };')
	ctx.execute('window.encodeURIComponent = function(_s){ };')
	ctx.execute(to_ecma5_function(f))
	return ctx



import lxml.html
from typing import Union, List
from typing import TYPE_CHECKING

from .atfile import AternosFile, FileType
if TYPE_CHECKING:
	from .atserver import AternosServer

class FileManager:

	def __init__(self, atserv:'AternosServer') -> None:

		self.atserv = atserv

	def listdir(self, path:str='') -> List[AternosFile]:

		path = path.lstrip('/')
		filesreq = self.atserv.atserver_request(
			f'https://aternos.org/files/{path}', 'GET'
		)
		filestree = lxml.html.fromstring(filesreq.content)
		fileslist = filestree.xpath('//div[contains(concat(" ",normalize-space(@class)," ")," file ")]')

		files = []
		for f in fileslist:

			ftype_raw = f.xpath('@data-type')[0]
			ftype = FileType.file \
				if ftype_raw == 'file' \
				else FileType.directory

			fsize_raw = f.xpath('./div[@class="filesize"]')
			fsize = 0
			if len(fsize_raw) > 0:

				fsize_text = fsize_raw[0].text.strip()
				fsize_num = fsize_text[:fsize_text.rfind(' ')]
				fsize_msr = fsize_text[fsize_text.rfind(' ')+1:]

				try:
					fsize = self.convert_size(float(fsize_num), fsize_msr)
				except ValueError:
					fsize = -1

			fullpath = f.xpath('@data-path')[0]
			filepath = fullpath[:fullpath.rfind('/')]
			filename = fullpath[fullpath.rfind('/'):]
			files.append(
				AternosFile(
					self.atserv,
					filepath, filename,
					ftype, fsize
				)
			)

		return files

	def convert_size(self, num:Union[int,float], measure:str) -> float:

		measure_match = {
			'B': 1,
			'kB': 1000,
			'MB': 1000000,
			'GB': 1000000000
		}
		try:
			return num * measure_match[measure]
		except KeyError:
			return -1

	def get_file(self, path:str) -> Union[AternosFile,None]:

		filepath = path[:path.rfind('/')]
		filename = path[path.rfind('/'):]

		filedir = self.listdir(filepath)
		for file in filedir:
			if file.name == filename:
				return file

		return None

	def dl_file(self, path:str) -> bytes:

		file = self.atserv.atserver_request(
			f'https://aternos.org/panel/ajax/files/download.php?' + \
			f'file={path.replace("/","%2F")}',
			'GET'
		)

		return file.content

	def dl_world(self, world:str='world') -> bytes:

		world = self.atserv.atserver_request(
			f'https://aternos.org/panel/ajax/worlds/download.php?' + \
			f'world={world.replace("/","%2F")}',
			'GET'
		)

		return world.content



import enum
import lxml.html
from typing import List, Union
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from .atserver import AternosServer

class Lists(enum.Enum):

	whl = 'whitelist'
	ops = 'ops'
	ban = 'banned-players'
	ips = 'banned-ips'

class PlayersList:

	def __init__(self, lst:Union[str,Lists], atserv:'AternosServer') -> None:

		self.atserv = atserv
		self.lst = Lists(lst)
		self.players = []
		self.parsed = False

	def list_players(self, cache:bool=True) -> List[str]:

		if cache and self.parsed:
			return self.players

		listreq = self.atserv.atserver_request(
			f'https://aternos.org/players/{self.lst.value}',
			'GET'
		)
		listtree = lxml.html.fromstring(listreq.content)
		items = listtree.xpath(
			'//div[@class="list-item"]'
		)

		result = []
		for i in items:
			name = i.xpath('./div[@class="list-name"]')
			result.append(name[0].text.strip())
		
		self.players = result
		self.parsed = True
		return result

	def add(self, name:str) -> None:

		self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/players/add.php',
			'POST', data={
				'list': self.lst.value,
				'name': name
			}, sendtoken=True
		)

		self.players.append(name)

	def remove(self, name:str) -> None:

		self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/players/remove.php',
			'POST', data={
				'list': self.lst.value,
				'name': name
			}, sendtoken=True
		)

		for i, j in enumerate(self.players):
			if j == name:
				del self.players[i]



import re
import random
import logging
from requests import Response
from cloudscraper import CloudScraper
from typing import Optional, Union, Dict

from . import atjsparse
from .aterrors import CredentialsError, CloudflareError

REQUA = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.84 Safari/537.36 OPR/85.0.4341.47'
REQHEADERS = {
	'Host': 'aternos.org',
	'User-Agent': REQUA,
	'Sec-Ch-Ua': '" Not A;Brand";v="99", "Chromium";v="100", "Opera";v="86"',
	'Sec-Ch-Ua-Mobile': '?0',
	'Sec-Ch-Ua-Platform': '"Linux"',
	'Sec-Fetch-Dest': 'document',
	'Sec-Fetch-Mode': 'navigate',
	'Sec-Fetch-Site': 'same-origin',
	'Sec-Fetch-User': '?1',
	'Upgrade-Insecure-Requests': '1'
}

class AternosConnect:

	def __init__(self) -> None:

		self.session = CloudScraper()
		self.atsession = ''

	def parse_token(self) -> str:

		loginpage = self.request_cloudflare(
			f'https://aternos.org/go/', 'GET'
		).content

		# Using the standard string methods
		# instead of the expensive xml parsing
		head = b'<head>'
		headtag = loginpage.find(head)
		headend = loginpage.find(b'</head>', headtag + len(head))

		# Some checks
		if headtag < 0 or headend < 0:
			pagehead = loginpage
			raise RuntimeWarning('Unable to find <head> tag, parsing the whole page')

		# Extracting <head> content
		headtag = headtag + len(head)
		pagehead = loginpage[headtag:headend]

		try:
			text = pagehead.decode('utf-8', 'replace')
			js_code = re.findall(r'\(\(\)(.*?)\)\(\);', text)
			token_func = js_code[1] if len(js_code) > 1 else js_code[0]

			ctx = atjsparse.exec(token_func)
			self.token = ctx.window['AJAX_TOKEN']

		except (IndexError, TypeError):
			raise CredentialsError(
				'Unable to parse TOKEN from the page'
			)

		return self.token

	def generate_sec(self) -> str:

		randkey = self.generate_aternos_rand()
		randval = self.generate_aternos_rand()
		self.sec = f'{randkey}:{randval}'
		self.session.cookies.set(
			f'ATERNOS_SEC_{randkey}', randval,
			domain='aternos.org'
		)

		return self.sec

	def generate_aternos_rand(self, randlen:int=16) -> str:

		# a list with randlen+1 empty strings:
		# generate a string with spaces,
		# then split it by space
		rand_arr = (' ' * (randlen+1)).split(' ')

		rand = random.random()
		rand_alphanum = self.convert_num(rand, 36) + ('0' * 17)

		return (rand_alphanum[:18].join(rand_arr)[:randlen])

	def convert_num(
		self, num:Union[int,float,str],
		base:int, frombase:int=10) -> str:

		if isinstance(num, str):
			num = int(num, frombase)

		if isinstance(num, float):
			sliced = str(num)[2:]
			num = int(sliced)

		symbols = '0123456789abcdefghijklmnopqrstuvwxyz'
		basesym = symbols[:base]
		result = ''
		while num > 0:
			rem = num % base
			result = str(basesym[rem]) + result
			num //= base
		return result

	def add_headers(self, headers:Optional[Dict[str,str]]=None) -> Dict[str,str]:

		headers = headers or {}
		headers.update(REQHEADERS)
		return headers

	def request_cloudflare(
		self, url:str, method:str,
		params:Optional[dict]=None, data:Optional[dict]=None,
		headers:Optional[dict]=None, reqcookies:Optional[dict]=None,
		sendtoken:bool=False, redirect:bool=True, retry:int=0) -> Response:

		if retry > 3:
			raise CloudflareError('Unable to bypass Cloudflare protection')

		try:
			self.atsession = self.session.cookies['ATERNOS_SESSION']
		except KeyError:
			pass

		if method not in ('GET', 'POST'):
			raise NotImplementedError('Only GET and POST are available')

		params = params or {}
		data = data or {}
		reqcookies = reqcookies or {}
		headers = self.add_headers(headers)

		if sendtoken:
			params['TOKEN'] = self.token
			params['SEC'] = self.sec
			headers['X-Requested-With'] = 'XMLHttpRequest'

		# requests.cookies.CookieConflictError bugfix
		reqcookies['ATERNOS_SESSION'] = self.atsession
		del self.session.cookies['ATERNOS_SESSION']

		logging.debug(f'Requesting({method})' + url)
		logging.debug('headers=' + str(headers))
		logging.debug('params=' + str(params))
		logging.debug('data=' + str(data))
		logging.debug('req-cookies=' + str(reqcookies))
		logging.debug('session-cookies=' + str(self.session.cookies))
		
		if method == 'POST':
			req = self.session.post(
				url, data=data, params=params,
				headers=headers, cookies=reqcookies,
				allow_redirects=redirect
			)
		else:
			req = self.session.get(
				url, params={**params, **data},
				headers=headers, cookies=reqcookies,
				allow_redirects=redirect
			)
		
		if '<title>Please Wait... | Cloudflare</title>' in req.text:
			logging.info('Retrying to bypass Cloudflare')
			self.request_cloudflare(
				url, method,
				params, data,
				headers, reqcookies,
				sendtoken, redirect,
				retry - 1
			)
		
		logging.info(
			f'{method} completed with {req.status_code} status'
		)

		return req



class AternosError(Exception):
	pass

class CloudflareError(AternosError):
	pass

class CredentialsError(AternosError):
	pass

class ServerError(AternosError):
	pass

class FileError(AternosError):
	pass



import enum
import json
from requests import Response
from typing import Optional

from .atconnect import AternosConnect
from .aterrors import ServerError
from .atfm import FileManager
from .atconf import AternosConfig
from .atplayers import PlayersList
from .atwss import AternosWss

class Edition(enum.IntEnum):
	java = 0
	bedrock = 1

class Status(enum.IntEnum):
	off = 0
	on = 1
	starting = 2
	shutdown = 3
	unknown = 6
	error = 7
	confirm = 10

class AternosServer:

	def __init__(
		self, servid:str,
		atconn:AternosConnect,
		reqinfo:bool=True) -> None:

		self.servid = servid
		self.atconn = atconn
		if reqinfo:
			self.fetch()
	
	def fetch(self) -> None:

		servreq = self.atserver_request(
			'https://aternos.org/panel/ajax/status.php',
			'GET', sendtoken=True
		)
		self._info = json.loads(servreq.content)

	def wss(self, autoconfirm:bool=False) -> AternosWss:

		return AternosWss(self, autoconfirm)

	def start(self, headstart:bool=False, accepteula:bool=True) -> None:

		startreq = self.atserver_request(
			'https://aternos.org/panel/ajax/start.php',
			'GET', params={'headstart': int(headstart)},
			sendtoken=True
		)
		startresult = startreq.json()

		if startresult['success']:
			return
		error = startresult['error']

		if error == 'eula' and accepteula:
			self.eula()
			self.start(accepteula=False)

		elif error == 'eula':
			raise ServerError(
				'EULA was not accepted. Use start(accepteula=True)'
			)

		elif error == 'already':
			raise ServerError(
				'Server is already running'
			)

		elif error == 'wrongversion':
			raise ServerError(
				'Incorrect software version installed'
			)

		elif error == 'file':
			raise ServerError(
				'File server is unavailbale, view https://status.aternos.gmbh'
			)

		elif error == 'size':
			raise ServerError(
				f'Available storage size is 4GB, ' + \
				f'your server used: {startresult["size"]}'
			)

		else:
			raise ServerError(
				f'Unable to start server, code: {error}'
			)

	def confirm(self) -> None:

		self.atserver_request(
			'https://aternos.org/panel/ajax/confirm.php',
			'GET', sendtoken=True
		)

	def stop(self) -> None:

		self.atserver_request(
			'https://aternos.org/panel/ajax/stop.php',
			'GET', sendtoken=True
		)

	def cancel(self) -> None:

		self.atserver_request(
			'https://aternos.org/panel/ajax/cancel.php',
			'GET', sendtoken=True
		)

	def restart(self) -> None:

		self.atserver_request(
			'https://aternos.org/panel/ajax/restart.php',
			'GET', sendtoken=True
		)

	def eula(self) -> None:

		self.atserver_request(
			'https://aternos.org/panel/ajax/eula.php',
			'GET', sendtoken=True
		)

	def files(self) -> FileManager:

		return FileManager(self)

	def config(self) -> AternosConfig:

		return AternosConfig(self)

	def players(self, lst:str) -> PlayersList:

		return PlayersList(lst, self)

	def atserver_request(
		self, url:str, method:int,
		params:Optional[dict]=None,
		data:Optional[dict]=None,
		headers:Optional[dict]=None,
		sendtoken:bool=False) -> Response:

		return self.atconn.request_cloudflare(
			url=url, method=method,
			params=params, data=data,
			headers=headers,
			reqcookies={
				'ATERNOS_SERVER': self.servid
			},
			sendtoken=sendtoken
		)

	@property
	def subdomain(self) -> str:
		atdomain = self.domain
		return atdomain[:atdomain.find('.')]

	@subdomain.setter
	def subdomain(self, value:str) -> None:
		self.atserver_request(
			'https://aternos.org/panel/ajax/options/subdomain.php',
			'GET', params={'subdomain': value},
			sendtoken=True
		)

	@property
	def motd(self) -> str:
		return self._info['motd']

	@motd.setter
	def motd(self, value:str) -> None:
		self.atserver_request(
			'https://aternos.org/panel/ajax/options/motd.php',
			'POST', data={'motd': value},
			sendtoken=True
		)

	@property
	def address(self) -> str:
		return f'{self.domain}:{self.port}'
	
	@property
	def domain(self) -> str:
		return self._info['ip']

	@property
	def port(self) -> int:
		return self._info['port']

	@property
	def edition(self) -> int:
		soft_type = self._info['bedrock']
		return int(soft_type)

	@property
	def software(self) -> str:
		return self._info['software']
	
	@property
	def version(self) -> str:
		return self._info['version']
	
	@property
	def status(self) -> str:
		return self._info['class']

	@property
	def status_num(self) -> int:
		return int(self._info['status'])

	@property
	def ram(self) -> int:
		return int(self._info['ram'])



import enum
import json
import asyncio
import logging
import websockets
from typing import Union, Any, Dict, Callable, Coroutine
from typing import TYPE_CHECKING

from .atconnect import REQUA
if TYPE_CHECKING:
	from .atserver import AternosServer

class Streams(enum.Enum):

	status = (0,None)
	queue = (1,None)
	console = (2,'console')
	ram = (3,'heap')
	tps = (4,'tick')

	def __init__(self, num:int, stream:str):
		self.num = num
		self.stream = stream

class AternosWss:

	def __init__(self, atserv:'AternosServer', autoconfirm:bool=False) -> None:
		
		self.atserv = atserv
		self.cookies = atserv.atconn.session.cookies
		self.session = self.cookies['ATERNOS_SESSION']
		self.servid = atserv.servid
		self.recv = {}
		self.autoconfirm = autoconfirm
		self.confirmed = False
	
	async def confirm(self) -> None:

		self.atserv.confirm()

	def wssreceiver(self, stream:Streams) -> Callable[[Callable[[Any],Coroutine[Any,Any,None]]],Any]:
		def decorator(func:Callable[[Any],Coroutine[Any,Any,None]]) -> None:
			self.recv[stream] = func
		return decorator

	async def connect(self) -> None:
		
		headers = [
			('Host', 'aternos.org'),
			('User-Agent', REQUA),
			(
				'Cookie',
				f'ATERNOS_SESSION={self.session}; ' + \
				f'ATERNOS_SERVER={self.servid}'
			)
		]
		self.socket = await websockets.connect(
			'wss://aternos.org/hermes/',
			origin='https://aternos.org',
			extra_headers=headers
		)

		@self.wssreceiver(Streams.status)
		async def confirmfunc(msg):
			# Autoconfirm
			if not self.autoconfirm:
				return
			if msg['class'] == 'queueing' \
			and msg['queue']['pending'] == 'pending'\
			and not self.confirmed:
				self.confirm()
		
		@self.wssreceiver(Streams.status)
		async def streamsfunc(msg):
			if msg['status'] == 2:
				# Automatically start streams
				for strm in self.recv:
					if not isinstance(strm,Streams):
						continue
					if strm.stream:
						logging.debug(f'Enabling {strm.stream} stream')
						await self.send({
							'stream': strm.stream,
							'type': 'start'
						})
		
		await self.wssworker()

	async def close(self) -> None:
		
		await self.socket.close()
		del self.socket

	async def send(self, obj:Union[Dict[str, Any],str]) -> None:

		if isinstance(obj, dict):
			obj = json.dumps(obj)

		await self.socket.send(obj)

	async def wssworker(self) -> None:

		keep = asyncio.create_task(self.keepalive())
		msgs = asyncio.create_task(self.receiver())
		await keep
		await msgs

	async def keepalive(self) -> None:

		while True:
			await asyncio.sleep(49)
			await self.socket.send('{"type":"\u2764"}')

	async def receiver(self) -> None:

		while True:
			data = await self.socket.recv()
			obj = json.loads(data)
			msgtype = -1
			
			if obj['type'] == 'line':
				msgtype = Streams.console
				msg = obj['data'].strip('\r\n ')

			elif obj['type'] == 'heap':
				msgtype = Streams.ram
				msg = int(obj['data']['usage'])

			elif obj['type'] == 'tick':
				msgtype = Streams.tps
				ticks = 1000 / obj['data']['averageTickTime']
				msg = 20 if ticks > 20 else ticks

			elif obj['type'] == 'status':
				msgtype = Streams.status
				msg = json.loads(obj['message'])

			if msgtype in self.recv:
				t = asyncio.create_task(
					self.recv[msgtype](msg)
				)
				await t



import enum
import lxml.html
from typing import Union
from typing import TYPE_CHECKING

from .aterrors import FileError

if TYPE_CHECKING:
	from .atserver import AternosServer

class FileType(enum.IntEnum):
	file = 0
	directory = 1

class AternosFile:

	def __init__(
		self, atserv:'AternosServer',
		path:str, name:str, ftype:int=FileType.file,
		size:Union[int,float]=0) -> None:

		self.atserv = atserv
		self._path = path.lstrip('/')
		self._name = name
		self._full = path + name
		self._ftype = ftype
		self._size = float(size)

	def delete(self) -> None:

		self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/delete.php',
			'POST', data={'file': self._full},
			sendtoken=True
		)

	def get_content(self) -> bytes:

		file = self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/files/download.php',
			'GET', params={
				'file': self._full
			}
		)
		if file.content == b'{"success":false}':
			raise FileError('Unable to download the file. Try to get text')
		return file.content

	def set_content(self, value:bytes) -> None:

		self.atserv.atserver_request(
			f'https://aternos.org/panel/ajax/save.php',
			'POST', data={
				'file': self._full,
				'content': value
			}, sendtoken=True
		)

	def get_text(self) -> str:

		editor = self.atserv.atserver_request(
			f'https://aternos.org/files/{self._full.lstrip("/")}', 'GET'
		)
		edittree = lxml.html.fromstring(editor.content)
		
		editblock = edittree.xpath('//div[@id="editor"]')[0]
		return editblock.text_content()

	def set_text(self, value:str) -> None:
		
		self.set_content(value.encode('utf-8'))

	@property
	def path(self):
		return self._path

	@property
	def name(self) -> str:
		return self._name
	
	@property
	def full(self) -> str:
		return self._full

	@property
	def is_dir(self) -> bool:
		if self._ftype == FileType.directory:
			return True
		return False

	@property
	def is_file(self) -> bool:
		if self._ftype == FileType.file:
			return True
		return False
	
	@property
	def size(self) -> float:
		return self._size



import os
import re
import hashlib
import lxml.html
from typing import List

from .atserver import AternosServer
from .atconnect import AternosConnect
from .aterrors import CredentialsError

__all__ = ['Client', 'atconf', 'atconnect', 'aterrors', 'atfile', 'atfm', 'atjsparse', 'atplayers', 'atserver', 'atwss']

class Client:

	def __init__(self, atconn:atconnect.AternosConnect) -> None:

		self.atconn = atconn

	@classmethod
	def from_hashed(cls, username:str, md5:str):

		atconn = AternosConnect()
		atconn.parse_token()
		atconn.generate_sec()

		credentials = {
			'user': username,
			'password': md5
		}

		loginreq = atconn.request_cloudflare(
			f'https://aternos.org/panel/ajax/account/login.php',
			'POST', data=credentials, sendtoken=True
		)

		if 'ATERNOS_SESSION' not in loginreq.cookies:
			raise CredentialsError(
				'Check your username and password'
			)

		return cls(atconn)

	@classmethod
	def from_credentials(cls, username:str, password:str):

		md5 = Client.md5encode(password)
		return cls.from_hashed(username, md5)

	@classmethod
	def from_session(cls, session:str):
		
		atconn = AternosConnect()
		atconn.session.cookies['ATERNOS_SESSION'] = session
		atconn.parse_token()
		atconn.generate_sec()

		return cls(atconn)
	
	@classmethod
	def restore_session(cls, file:str='~/.aternos'):

		file = os.path.expanduser(file)
		with open(file, 'rt') as f:
			session = f.read().strip()
		return cls.from_session(session)
	
	@staticmethod
	def md5encode(passwd:str) -> str:

		encoded = hashlib.md5(passwd.encode('utf-8'))
		return encoded.hexdigest().lower()
	
	def save_session(self, file:str='~/.aternos') -> None:

		file = os.path.expanduser(file)
		with open(file, 'wt') as f:
			f.write(self.atconn.atsession)

	def list_servers(self) -> List[AternosServer]:

		serverspage = self.atconn.request_cloudflare(
			'https://aternos.org/servers/', 'GET'
		)
		serverstree = lxml.html.fromstring(serverspage.content)
		serverslist = serverstree.xpath('//div[contains(@class,"servers ")]/div')

		servers = []
		for server in serverslist:
			servid = server.xpath('./div[@class="server-body"]/@data-id')[0]
			servers.append(AternosServer(servid, self.atconn))

		return servers
	
	def get_server(self, servid:str) -> AternosServer:

		return AternosServer(servid, self.atconn)
	
	def change_username(self, value:str) -> None:

		self.atconn.request_cloudflare(
			'https://aternos.org/panel/ajax/account/username.php',
			'POST', data={'username': value}
		)
	
	def change_email(self, value:str) -> None:

		email = re.compile(r'^[A-Za-z0-9\-_+.]+@[A-Za-z0-9\-_+.]+\.[A-Za-z0-9\-]+$|^$')
		if not email.match(value):
			raise ValueError('Invalid e-mail!')

		self.atconn.request_cloudflare(
			'https://aternos.org/panel/ajax/account/email.php',
			'POST', data={'email': value}
		)
	
	def change_password(self, old:str, new:str) -> None:

		old = Client.md5encode(old)
		new = Client.md5encode(new)
		self.atconn.request_cloudflare(
			'https://aternos.org/panel/ajax/account/password.php',
			'POST', data={
				'oldpassword': old,
				'newpassword': new
			}
		)



import enum
import re
import lxml.html
from typing import Any, Dict, List, Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from .atserver import AternosServer

#
# server.options
class ServerOpts(enum.Enum):
	players = 'max-players'
	gm = 'gamemode'
	difficulty = 'difficulty'
	whl = 'white-list'
	online = 'online-mode'
	pvp = 'pvp'
	cmdblock = 'enable-command-block'
	flight = 'allow-flight'
	animals = 'spawn-animals'
	monsters = 'spawn-monsters'
	villagers = 'spawn-npcs'
	nether = 'allow-nether'
	forcegm = 'force-gamemode'
	spawnlock = 'spawn-protection'
	cmds = 'allow-cheats'
	packreq = 'require-resource-pack'
	pack = 'resource-pack'

DAT_PREFIX = 'Data:'
DAT_GR_PREFIX = 'Data:GameRules:'

# level.dat
class WorldOpts(enum.Enum):
	seed12 = 'randomseed'
	seed = 'seed'
	hardcore = 'hardcore'
	difficulty = 'Difficulty'

# /gamerule
class WorldRules(enum.Enum):
	advs = 'announceAdvancements'
	univanger = 'universalAnger'
	cmdout = 'commandBlockOutput'
	elytra = 'disableElytraMovementCheck'
	raids = 'disableRaids'
	daynight = 'doDaylightCycle'
	entdrop = 'doEntityDrops'
	fire = 'doFireTick'
	phantoms = 'doInsomnia'
	immrespawn = 'doImmediateRespawn'
	limitcraft = 'doLimitedCrafting'
	mobloot = 'doMobLoot'
	mobs = 'doMobSpawning'
	patrols = 'doPatrolSpawning'
	blockdrop = 'doTileDrops'
	traders = 'doTraderSpawning'
	weather = 'doWeatherCycle'
	drowndmg = 'drowningDamage'
	falldmg = 'fallDamage'
	firedmg = 'fireDamage'
	snowdmg = 'freezeDamage'
	forgive = 'forgiveDeadPlayers'
	keepinv = 'keepInventory'
	deathmsg = 'showDeathMessages'
	admincmdlog = 'logAdminCommands'
	cmdlen = 'maxCommandChainLength'
	entcram = 'maxEntityCramming'
	mobgrief = 'mobGriefing'
	regen = 'naturalRegeneration'
	sleeppct = 'playersSleepingPercentage'
	rndtick = 'randomTickSpeed'
	spawnradius = 'spawnRadius'
	reducedf3 = 'reducedDebugInfo'
	spectchunkgen = 'spectatorsGenerateChunks'
	cmdfb = 'sendCommandFeedback'

DAT_TYPE_WORLD = 0
DAT_TYPE_GR = 1

class Gamemode(enum.IntEnum):
	survival = 0
	creative = 1
	adventure = 2
	spectator = 3

class Difficulty(enum.IntEnum):
	peaceful = 0
	easy = 1
	normal = 2
	hard = 3

#
# jre types for set_java
javatype = {
	'jdk': 'openjdk:{ver}',
	'openj9-1': 'adoptopenjdk:{ver}-jre-openj9-bionic',
	'openj9-2': 'ibm-semeru-runtimes:open-{ver}-jre'
}
# checking java version format
javacheck = re.compile(
	''.join(
		list(
			map(
				# create a regexp for each jre type,
				# e.g.: (^openjdk:\d+$)|
				lambda i: '(^' + javatype[i].format(ver=r'\d+') + '$)|',
				javatype
			)
		)
	).rstrip('|')
)
# checking timezone format
tzcheck = re.compile(r'(^[A-Z]\w+\/[A-Z]\w+$)|^UTC$')
# options types converting
convert = {
	'config-option-number': int,
	'config-option-select': int,
	'config-option-toggle': bool
}

# MAIN CLASS
class AternosConfig:

	def __init__(self, atserv:'AternosServer') -> None:

		self.atserv = atserv

	def get_timezone(self) -> str:

		optreq = self.atserv.atserver_request(
			'https://aternos.org/options', 'GET'
		)
		opttree = lxml.html.fromstring(optreq)

		tzopt = opttree.xpath('//div[@class="options-other-input timezone-switch"]')[0]
		tztext = tzopt.xpath('.//div[@class="option current"]')[0].text
		return tztext.strip()

	def set_timezone(self, value:str) -> None:

		matches_tz = tzcheck.search(value)
		if not matches_tz:
			raise ValueError('Timezone must match zoneinfo format: Area/Location')

		self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/timezone.php',
			'POST', data={'timezone': value},
			sendtoken=True
		)

	def get_java(self) -> str:

		optreq = self.atserv.atserver_request(
			'https://aternos.org/options', 'GET'
		)
		opttree = lxml.html.fromstring(optreq)
		imgopt = opttree.xpath('//div[@class="options-other-input image-switch"]')[0]
		imgver = imgopt.xpath('.//div[@class="option current"]/@data-value')[0]
		return imgver

	def set_java(self, value:str) -> None:

		matches_jdkver = javacheck.search(value)
		if not matches_jdkver:
			raise ValueError('Incorrect Java image version format!')

		self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/image.php',
			'POST', data={'image': value},
			sendtoken=True
		)

	#
	# server.properties
	#
	def set_server_prop(self, option:str, value:Any) -> None:
		self.__set_prop(
			'/server.properties',
			option, value
		)

	def get_server_props(self, proptyping:bool=True) -> Dict[str,Any]:
		return self.__get_all_props('https://aternos.org/options', proptyping)

	def set_server_props(self, props:Dict[str,Any]) -> None:
		for key in props:
			self.set_server_prop(key, props[key])

	#
	# level.dat
	#
	def set_world_prop(
		self, option:str, value:Any,
		proptype:int, world:str='world') -> None:

		prefix = DAT_PREFIX
		if proptype == DAT_TYPE_GR:
			prefix = DAT_GR_PREFIX

		self.__set_prop(
			f'/{world}/level.dat',
			f'{prefix}{option}',
			value
		)

	def get_world_props(
		self, world:str='world',
		proptyping:bool=True) -> Dict[str,Any]:

		self.__get_all_props(
			f'https://aternos.org/files/{world}/level.dat',
			proptyping, [DAT_PREFIX, DAT_GR_PREFIX]
		)

	def set_world_props(self, props:Dict[str,Any]) -> None:
		for key in props:
			self.set_world_prop(key, props[key])

	#
	# helpers
	#
	def __set_prop(self, file:str, option:str, value:Any) -> None:

		self.atserv.atserver_request(
			'https://aternos.org/panel/ajax/config.php',
			'POST', data={
				'file': file,
				'option': option,
				'value': value
			}, sendtoken=True
		)

	def __get_all_props(
		self, url:str, proptyping:bool=True,
		prefixes:Optional[List[str]]=None) -> Dict[str,Any]:

		optreq = self.atserv.atserver_request(url, 'GET')
		opttree = lxml.html.fromstring(optreq.content)
		configs = opttree.xpath('//div[@class="config-options"]')

		for i, conf in enumerate(configs):
			opts = conf.xpath('/div[contains(@class,"config-option ")]')
			result = {}

			for opt in opts:
				key = opt.xpath('.//span[@class="config-option-output-key"]')[0].text
				value = opt.xpath('.//span[@class="config-option-output-value"]')[0].text

				if prefixes != None:
					key = f'{prefixes[i]}{key}'

				opttype = opt.xpath('/@class').split(' ')[1]
				if proptyping and opttype in convert:
					value = convert[opttype](value)

				result[key] = value

		return result



import setuptools

with open('README.md', 'rt') as readme:
	long_description = readme.read()

setuptools.setup(
	name='python-aternos',
	version='1.0.3',
	author='Chechkenev Andrey (@DarkCat09)',
	author_email='aacd0709@mail.ru',
	description='An unofficial Aternos API',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/DarkCat09/python-aternos',
	project_urls={
		'Homepage': 'https://codeberg.org/DarkCat09/python-aternos',
		'GitHub': 'https://github.com/DarkCat09/python-aternos',
		'Bug Tracker': 'https://github.com/DarkCat09/python-aternos/issues',
		'Bug Tracker 2': 'https://codeberg.org/DarkCat09/python-aternos/issues'
	},
	classifiers=[
		'Development Status :: 4 - Beta',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'License :: OSI Approved :: Apache Software License',
		'Operating System :: OS Independent',
		'Operating System :: Microsoft :: Windows',
		'Operating System :: POSIX :: Linux',
		'Operating System :: MacOS',
		'Operating System :: Other OS'
	],
	install_requires=[
		'lxml>=4.8.0',
		'cloudscraper>=1.2.58',
		'js2py>=0.71',
		'websockets>=10.1',
		'regex>=2022.3.15'
	],
	packages=['python_aternos'],
	python_requires=">=3.7",
)
